@@ -7,12 +7,14 @@ import collections.abc
 import copy
 import inspect
 import re
-from dataclasses import fields
+from abc import ABCMeta
+from dataclasses import dataclass, fields
 from enum import Enum, auto
 from typing import (
     Callable,
     Dict,
     Generic,
+    Iterator,
     List,
     Mapping,
     NoReturn,
@@ -51,11 +53,26 @@ _CallableT = TypeVar("_CallableT", bound="Callable", covariant=True)
 _BaseMatcherNodeSelfT = TypeVar("_BaseMatcherNodeSelfT", bound="BaseMatcherNode")
 _OtherNodeT = TypeVar("_OtherNodeT")
 _MetadataValueT = TypeVar("_MetadataValueT")
+_MatcherTypeT = TypeVar("_MatcherTypeT", bound=Type["BaseMatcherNode"])
+_OtherNodeMatcherTypeT = TypeVar(
+    "_OtherNodeMatcherTypeT", bound=Type["BaseMatcherNode"]
+)
 
 
 _METADATA_MISSING_SENTINEL = object()
 
 
+class AbstractBaseMatcherNodeMeta(ABCMeta):
+    """
+    Metaclass that all matcher nodes uses. Allows chaining 2 node type
+    together with an bitwise-or operator to produce an :class:`TypeOf`
+    matcher.
+    """
+
+    def __or__(self, node: Type["BaseMatcherNode"]) -> "TypeOf[Type[BaseMatcherNode]]":
+        return TypeOf(self, node)
+
+
 class BaseMatcherNode:
     """
     Base class that all concrete matchers subclass from. :class:`OneOf` and
@@ -103,6 +120,81 @@ def DoNotCare() -> DoNotCareSentinel:
     return DoNotCareSentinel.DEFAULT
 
 
+class TypeOf(Generic[_MatcherTypeT], BaseMatcherNode):
+    """
+    Matcher that matches any one of the given types. Useful when you want to work
+    with trees where a common property might belong to more than a single type.
+
+    For example, if you want either a binary operation or a boolean operation
+    where the left side has a name ``foo``::
+
+        m.TypeOf(m.BinaryOperation, m.BooleanOperation)(left = m.Name("foo"))
+
+    Or you could use the shorthand, like::
+
+        (m.BinaryOperation | m.BooleanOperation)(left = m.Name("foo"))
+
+    Also :class:`TypeOf` matchers can be used with initalizing in the default
+    state of other node matchers (without passing any extra patterns)::
+
+        m.Name | m.SimpleString
+
+    The will be equal to::
+
+        m.OneOf(m.Name(), m.SimpleString())
+    """
+
+    def __init__(self, *options: Union[_MatcherTypeT, "TypeOf[_MatcherTypeT]"]) -> None:
+        actual_options: List[_MatcherTypeT] = []
+        for option in options:
+            if isinstance(option, TypeOf):
+                if option.initalized:
+                    raise Exception(
+                        "Cannot chain an uninitalized TypeOf with an initalized one"
+                    )
+                actual_options.extend(option._raw_options)
+            else:
+                actual_options.append(option)
+
+        self._initalized = False
+        self._call_items: Tuple[Tuple[object, ...], Dict[str, object]] = ((), {})
+        self._raw_options: Tuple[_MatcherTypeT, ...] = tuple(actual_options)
+
+    @property
+    def initalized(self) -> bool:
+        return self._initalized
+
+    @property
+    def options(self) -> Iterator[BaseMatcherNode]:
+        for option in self._raw_options:
+            args, kwargs = self._call_items
+            matcher_pattern = option(*args, **kwargs)
+            yield matcher_pattern
+
+    def __call__(self, *args: object, **kwargs: object) -> BaseMatcherNode:
+        self._initalized = True
+        self._call_items = (args, kwargs)
+        return self
+
+    def __or__(
+        self, other: _OtherNodeMatcherTypeT
+    ) -> "TypeOf[Union[_MatcherTypeT, _OtherNodeMatcherTypeT]]":
+        return TypeOf[Union[_MatcherTypeT, _OtherNodeMatcherTypeT]](self, other)
+
+    def __and__(self, other: _OtherNodeMatcherTypeT) -> NoReturn:
+        left, right = type(self).__name__, other.__name__
+        raise TypeError(
+            f"TypeError: unsupported operand type(s) for &: {left!r} and {right!r}"
+        )
+
+    def __invert__(self) -> "AllOf[BaseMatcherNode]":
+        return AllOf(*map(DoesNotMatch, self.options))
+
+    def __repr__(self) -> str:
+        types = ", ".join(repr(option) for option in self._raw_options)
+        return f"TypeOf({types}, initalized = {self.initalized})"
+
+
 class OneOf(Generic[_MatcherT], BaseMatcherNode):
     """
     Matcher that matches any one of its options. Useful when you want to match
@@ -352,8 +444,10 @@ class _ExtractMatchingNode(Generic[_MatcherT]):
         # same node, or none of them. It makes more sense to move the SaveMatchedNode
         # up to wrap the AllOf.
         raise Exception(
-            "Cannot use AllOf with SavedMatchedNode children! Instead, you should "
-            "use SaveMatchedNode(AllOf(options...))."
+            (
+                "Cannot use AllOf with SavedMatchedNode children! Instead, you should "
+                + "use SaveMatchedNode(AllOf(options...))."
+            )
         )
 
     def __getattr__(self, key: str) -> object:
@@ -366,8 +460,10 @@ class _ExtractMatchingNode(Generic[_MatcherT]):
         # This doesn't make sense. We don't want to capture a node only if it
         # doesn't match, since this will never capture anything.
         raise Exception(
-            "Cannot invert a SaveMatchedNode. Instead you should wrap SaveMatchedNode "
-            "around your inversion itself"
+            (
+                "Cannot invert a SaveMatchedNode. Instead you should wrap SaveMatchedNode "
+                + "around your inversion itself"
+            )
         )
 
     def __repr__(self) -> str:
@@ -706,10 +802,10 @@ class AtLeastN(Generic[_MatcherT], _BaseWildcardNode):
         return self._matcher
 
     def __or__(self, other: object) -> NoReturn:
-        raise Exception(f"AtLeastN cannot be used in a OneOf matcher")
+        raise Exception("AtLeastN cannot be used in a OneOf matcher")
 
     def __and__(self, other: object) -> NoReturn:
-        raise Exception(f"AtLeastN cannot be used in an AllOf matcher")
+        raise Exception("AtLeastN cannot be used in an AllOf matcher")
 
     def __invert__(self) -> NoReturn:
         raise Exception("Cannot invert an AtLeastN matcher!")
@@ -808,10 +904,10 @@ class AtMostN(Generic[_MatcherT], _BaseWildcardNode):
         return self._matcher
 
     def __or__(self, other: object) -> NoReturn:
-        raise Exception(f"AtMostN cannot be used in a OneOf matcher")
+        raise Exception("AtMostN cannot be used in a OneOf matcher")
 
     def __and__(self, other: object) -> NoReturn:
-        raise Exception(f"AtMostN cannot be used in an AllOf matcher")
+        raise Exception("AtMostN cannot be used in an AllOf matcher")
 
     def __invert__(self) -> NoReturn:
         raise Exception("Cannot invert an AtMostN matcher!")
@@ -932,6 +1028,34 @@ def SaveMatchedNode(matcher: _OtherNodeT, name: str) -> _OtherNodeT:
     return cast(_OtherNodeT, _ExtractMatchingNode(matcher, name))
 
 
+def _matches_zero_nodes(
+    matcher: Union[
+        BaseMatcherNode,
+        _BaseWildcardNode,
+        MatchIfTrue[Callable[[object], bool]],
+        _BaseMetadataMatcher,
+        DoNotCareSentinel,
+    ]
+) -> bool:
+    if isinstance(matcher, AtLeastN) and matcher.n == 0:
+        return True
+    if isinstance(matcher, AtMostN):
+        return True
+    if isinstance(matcher, _ExtractMatchingNode):
+        return _matches_zero_nodes(matcher.matcher)
+    return False
+
+
+@dataclass(frozen=True)
+class _SequenceMatchesResult:
+    sequence_capture: Optional[
+        Dict[str, Union[libcst.CSTNode, Sequence[libcst.CSTNode]]]
+    ]
+    matched_nodes: Optional[
+        Union[libcst.CSTNode, MaybeSentinel, Sequence[libcst.CSTNode]]
+    ]
+
+
 def _sequence_matches(  # noqa: C901
     nodes: Sequence[Union[MaybeSentinel, libcst.CSTNode]],
     matchers: Sequence[
@@ -944,30 +1068,34 @@ def _sequence_matches(  # noqa: C901
         ]
     ],
     metadata_lookup: Callable[[meta.ProviderT, libcst.CSTNode], object],
-) -> Optional[Dict[str, Union[libcst.CSTNode, Sequence[libcst.CSTNode]]]]:
+) -> _SequenceMatchesResult:
     if not nodes and not matchers:
-        # Base case, empty lists are alwatys matches
-        return {}
+        # Base case, empty lists are always matches
+        return _SequenceMatchesResult({}, None)
     if not nodes and matchers:
         # Base case, we have one or more matcher that wasn't matched
-        return (
-            {}
-            if all(
-                (isinstance(m, AtLeastN) and m.n == 0) or isinstance(m, AtMostN)
-                for m in matchers
+        if all(_matches_zero_nodes(m) for m in matchers):
+            return _SequenceMatchesResult(
+                {m.name: () for m in matchers if isinstance(m, _ExtractMatchingNode)},
+                (),
             )
-            else None
-        )
+        else:
+            return _SequenceMatchesResult(None, None)
     if nodes and not matchers:
         # Base case, we have nodes left that don't match any matcher
-        return None
+        return _SequenceMatchesResult(None, None)
 
     # Recursive case, nodes and matchers LHS matches
     node = nodes[0]
     matcher = matchers[0]
     if isinstance(matcher, DoNotCareSentinel):
         # We don't care about the value for this node.
-        return _sequence_matches(nodes[1:], matchers[1:], metadata_lookup)
+        return _SequenceMatchesResult(
+            _sequence_matches(
+                nodes[1:], matchers[1:], metadata_lookup
+            ).sequence_capture,
+            node,
+        )
     elif isinstance(matcher, _BaseWildcardNode):
         if isinstance(matcher, AtMostN):
             if matcher.n > 0:
@@ -977,18 +1105,24 @@ def _sequence_matches(  # noqa: C901
                     nodes[0], matcher.matcher, metadata_lookup
                 )
                 if attribute_capture is not None:
-                    sequence_capture = _sequence_matches(
+                    result = _sequence_matches(
                         nodes[1:],
                         [AtMostN(matcher.matcher, n=matcher.n - 1), *matchers[1:]],
                         metadata_lookup,
                     )
-                    if sequence_capture is not None:
-                        return {**attribute_capture, **sequence_capture}
+                    if result.sequence_capture is not None:
+                        return _SequenceMatchesResult(
+                            {**attribute_capture, **result.sequence_capture},
+                            (node, *result.matched_nodes),
+                        )
             # Finally, assume that this does not match the current node.
             # Consume the matcher but not the node.
-            sequence_capture = _sequence_matches(nodes, matchers[1:], metadata_lookup)
-            if sequence_capture is not None:
-                return sequence_capture
+            return _SequenceMatchesResult(
+                _sequence_matches(
+                    nodes, matchers[1:], metadata_lookup
+                ).sequence_capture,
+                (),
+            )
         elif isinstance(matcher, AtLeastN):
             if matcher.n > 0:
                 # Only match if we can consume one of the matches, since we still
@@ -997,13 +1131,17 @@ def _sequence_matches(  # noqa: C901
                     nodes[0], matcher.matcher, metadata_lookup
                 )
                 if attribute_capture is not None:
-                    sequence_capture = _sequence_matches(
+                    result = _sequence_matches(
                         nodes[1:],
                         [AtLeastN(matcher.matcher, n=matcher.n - 1), *matchers[1:]],
                         metadata_lookup,
                     )
-                    if sequence_capture is not None:
-                        return {**attribute_capture, **sequence_capture}
+                    if result.sequence_capture is not None:
+                        return _SequenceMatchesResult(
+                            {**attribute_capture, **result.sequence_capture},
+                            (node, *result.matched_nodes),
+                        )
+                return _SequenceMatchesResult(None, None)
             else:
                 # First, assume that this does match a node (greedy).
                 # Consume one node since it matched this matcher.
@@ -1011,45 +1149,52 @@ def _sequence_matches(  # noqa: C901
                     nodes[0], matcher.matcher, metadata_lookup
                 )
                 if attribute_capture is not None:
-                    sequence_capture = _sequence_matches(
-                        nodes[1:], matchers, metadata_lookup
-                    )
-                    if sequence_capture is not None:
-                        return {**attribute_capture, **sequence_capture}
+                    result = _sequence_matches(nodes[1:], matchers, metadata_lookup)
+                    if result.sequence_capture is not None:
+                        return _SequenceMatchesResult(
+                            {**attribute_capture, **result.sequence_capture},
+                            (node, *result.matched_nodes),
+                        )
                 # Now, assume that this does not match the current node.
                 # Consume the matcher but not the node.
-                sequence_capture = _sequence_matches(
-                    nodes, matchers[1:], metadata_lookup
+                return _SequenceMatchesResult(
+                    _sequence_matches(
+                        nodes, matchers[1:], metadata_lookup
+                    ).sequence_capture,
+                    (),
                 )
-                if sequence_capture is not None:
-                    return sequence_capture
         else:
             # There are no other types of wildcard consumers, but we're making
             # pyre happy with that fact.
             raise Exception(f"Logic error unrecognized wildcard {type(matcher)}!")
     elif isinstance(matcher, _ExtractMatchingNode):
         # See if the raw matcher matches. If it does, capture the sequence we matched and store it.
-        sequence_capture = _sequence_matches(
+        result = _sequence_matches(
             nodes, [matcher.matcher, *matchers[1:]], metadata_lookup
         )
-        if sequence_capture is not None:
-            return {
-                # Our own match capture comes first, since we wnat to allow the same
-                # name later in the sequence to override us.
-                matcher.name: nodes,
-                **sequence_capture,
-            }
-        return None
+        if result.sequence_capture is not None:
+            return _SequenceMatchesResult(
+                {
+                    # Our own match capture comes first, since we wnat to allow the same
+                    # name later in the sequence to override us.
+                    matcher.name: result.matched_nodes,
+                    **result.sequence_capture,
+                },
+                result.matched_nodes,
+            )
+        return _SequenceMatchesResult(None, None)
 
     match_capture = _matches(node, matcher, metadata_lookup)
     if match_capture is not None:
         # These values match directly
-        sequence_capture = _sequence_matches(nodes[1:], matchers[1:], metadata_lookup)
-        if sequence_capture is not None:
-            return {**match_capture, **sequence_capture}
+        result = _sequence_matches(nodes[1:], matchers[1:], metadata_lookup)
+        if result.sequence_capture is not None:
+            return _SequenceMatchesResult(
+                {**match_capture, **result.sequence_capture}, node
+            )
 
     # Failed recursive case, no match
-    return None
+    return _SequenceMatchesResult(None, None)
 
 
 _AttributeValueT = Optional[Union[MaybeSentinel, libcst.CSTNode, str, bool]]
@@ -1110,9 +1255,9 @@ def _attribute_matches(  # noqa: C901
             for m in matcher.options:
                 if isinstance(m, collections.abc.Sequence):
                     # Should match the sequence of requested nodes
-                    sequence_capture = _sequence_matches(node, m, metadata_lookup)
-                    if sequence_capture is not None:
-                        return sequence_capture
+                    result = _sequence_matches(node, m, metadata_lookup)
+                    if result.sequence_capture is not None:
+                        return result.sequence_capture
                 elif isinstance(m, MatchIfTrue):
                     return {} if matcher.func(node) else None
         elif isinstance(matcher, AllOf):
@@ -1121,10 +1266,10 @@ def _attribute_matches(  # noqa: C901
             for m in matcher.options:
                 if isinstance(m, collections.abc.Sequence):
                     # Should match the sequence of requested nodes
-                    sequence_capture = _sequence_matches(node, m, metadata_lookup)
-                    if sequence_capture is None:
+                    result = _sequence_matches(node, m, metadata_lookup)
+                    if result.sequence_capture is None:
                         return None
-                    all_captures = {**all_captures, **sequence_capture}
+                    all_captures = {**all_captures, **result.sequence_capture}
                 elif isinstance(m, MatchIfTrue):
                     return {} if matcher.func(node) else None
                 else:
@@ -1150,7 +1295,7 @@ def _attribute_matches(  # noqa: C901
                     matcher,
                 ),
                 metadata_lookup,
-            )
+            ).sequence_capture
 
         # We exhausted our possibilities, there's no match
         return None
@@ -1334,7 +1479,7 @@ def _matches(
         return {} if isinstance(matcher, _InverseOf) else None
 
     # Now, evaluate the matcher node itself.
-    if isinstance(matcher, OneOf):
+    if isinstance(matcher, (OneOf, TypeOf)):
         for matcher in matcher.options:
             node_capture = _node_matches(node, matcher, metadata_lookup)
             if node_capture is not None:
