@@ -36,6 +36,25 @@ from libcst.metadata.expression_context_provider import (
 )
 
 
+_ASSIGNMENT_LIKE_NODES = (
+    cst.AnnAssign,
+    cst.AsName,
+    cst.Assign,
+    cst.AugAssign,
+    cst.ClassDef,
+    cst.CompFor,
+    cst.For,
+    cst.FunctionDef,
+    cst.Global,
+    cst.Import,
+    cst.ImportFrom,
+    cst.NamedExpr,
+    cst.Nonlocal,
+    cst.Parameters,
+    cst.WithItem,
+)
+
+
 @add_slots
 @dataclass(frozen=False)
 class Access:
@@ -65,13 +84,20 @@ class Access:
 
     is_annotation: bool
 
-    __assignments: Set["BaseAssignment"]
+    is_type_hint: bool
 
-    def __init__(self, node: cst.Name, scope: "Scope", is_annotation: bool) -> None:
+    __assignments: Set["BaseAssignment"]
+    __index: int
+
+    def __init__(
+        self, node: cst.Name, scope: "Scope", is_annotation: bool, is_type_hint: bool
+    ) -> None:
         self.node = node
         self.scope = scope
         self.is_annotation = is_annotation
+        self.is_type_hint = is_type_hint
         self.__assignments = set()
+        self.__index = scope._assignment_count
 
     def __hash__(self) -> int:
         return id(self)
@@ -81,11 +107,25 @@ class Access:
         """Return all assignments of the access."""
         return self.__assignments
 
-    def record_assignment(self, assignment: "BaseAssignment") -> None:
-        self.__assignments.add(assignment)
+    @property
+    def _index(self) -> int:
+        return self.__index
 
-    def record_assignments(self, assignments: Set["BaseAssignment"]) -> None:
-        self.__assignments |= assignments
+    def record_assignment(self, assignment: "BaseAssignment") -> None:
+        if assignment.scope != self.scope or assignment._index < self.__index:
+            self.__assignments.add(assignment)
+
+    def record_assignments(self, name: str) -> None:
+        assignments = self.scope[name]
+        # filter out assignments that happened later than this access
+        previous_assignments = {
+            assignment
+            for assignment in assignments
+            if assignment.scope != self.scope or assignment._index < self.__index
+        }
+        if not previous_assignments and assignments:
+            previous_assignments = self.scope.parent[name]
+        self.__assignments |= previous_assignments
 
 
 class BaseAssignment(abc.ABC):
@@ -104,10 +144,22 @@ class BaseAssignment(abc.ABC):
         self.__accesses = set()
 
     def record_access(self, access: Access) -> None:
-        self.__accesses.add(access)
+        if access.scope != self.scope or self._index < access._index:
+            self.__accesses.add(access)
 
     def record_accesses(self, accesses: Set[Access]) -> None:
-        self.__accesses |= accesses
+        later_accesses = {
+            access
+            for access in accesses
+            if access.scope != self.scope or self._index < access._index
+        }
+        self.__accesses |= later_accesses
+        earlier_accesses = accesses - later_accesses
+        if earlier_accesses and self.scope.parent != self.scope:
+            # Accesses "earlier" than the relevant assignment should be attached
+            # to assignments of the same name in the parent
+            for shadowed_assignment in self.scope.parent[self.name]:
+                shadowed_assignment.record_accesses(earlier_accesses)
 
     @property
     def references(self) -> Collection[Access]:
@@ -118,6 +170,11 @@ class BaseAssignment(abc.ABC):
     def __hash__(self) -> int:
         return id(self)
 
+    @property
+    def _index(self) -> int:
+        """Return an integer that represents the order of assignments in `scope`"""
+        return -1
+
 
 class Assignment(BaseAssignment):
     """An assignment records the name, CSTNode and its accesses."""
@@ -125,11 +182,19 @@ class Assignment(BaseAssignment):
     #: The node of assignment, it could be a :class:`~libcst.Import`, :class:`~libcst.ImportFrom`,
     #: :class:`~libcst.Name`, :class:`~libcst.FunctionDef`, or :class:`~libcst.ClassDef`.
     node: cst.CSTNode
+    __index: int
 
-    def __init__(self, name: str, scope: "Scope", node: cst.CSTNode) -> None:
+    def __init__(
+        self, name: str, scope: "Scope", node: cst.CSTNode, index: int
+    ) -> None:
         self.node = node
+        self.__index = index
         super().__init__(name, scope)
 
+    @property
+    def _index(self) -> int:
+        return self.__index
+
 
 # even though we don't override the constructor.
 class BuiltinAssignment(BaseAssignment):
@@ -313,6 +378,7 @@ class Scope(abc.ABC):
     globals: "GlobalScope"
     _assignments: MutableMapping[str, Set[BaseAssignment]]
     _accesses: MutableMapping[str, Set[Access]]
+    _assignment_count: int
 
     def __init__(self, parent: "Scope") -> None:
         super().__init__()
@@ -320,9 +386,12 @@ class Scope(abc.ABC):
         self.globals = parent.globals
         self._assignments = defaultdict(set)
         self._accesses = defaultdict(set)
+        self._assignment_count = 0
 
     def record_assignment(self, name: str, node: cst.CSTNode) -> None:
-        self._assignments[name].add(Assignment(name=name, scope=self, node=node))
+        self._assignments[name].add(
+            Assignment(name=name, scope=self, node=node, index=self._assignment_count)
+        )
 
     def record_access(self, name: str, access: Access) -> None:
         self._accesses[name].add(access)
@@ -646,7 +715,9 @@ class ScopeVisitor(cst.CSTVisitor):
         self.__in_annotation: Set[
             Union[cst.Call, cst.Annotation, cst.Subscript]
         ] = set()
+        self.__in_type_hint: Set[Union[cst.Call, cst.Annotation, cst.Subscript]] = set()
         self.__in_ignored_subscript: Set[cst.Subscript] = set()
+        self.__ignore_annotation: int = 0
 
     @contextmanager
     def _new_scope(
@@ -702,18 +773,24 @@ class ScopeVisitor(cst.CSTVisitor):
 
     def visit_Call(self, node: cst.Call) -> Optional[bool]:
         self.__top_level_attribute_stack.append(None)
-        qnames = self.scope.get_qualified_names_for(node)
-        if any(qn.name in {"typing.NewType", "typing.TypeVar"} for qn in qnames):
+        qnames = {qn.name for qn in self.scope.get_qualified_names_for(node)}
+        if "typing.NewType" in qnames or "typing.TypeVar" in qnames:
             node.func.visit(self)
-            self.__in_annotation.add(node)
+            self.__in_type_hint.add(node)
             for arg in node.args[1:]:
                 arg.visit(self)
             return False
+        if "typing.cast" in qnames:
+            node.func.visit(self)
+            self.__in_type_hint.add(node)
+            if len(node.args) > 0:
+                node.args[0].visit(self)
+            return False
         return True
 
     def leave_Call(self, original_node: cst.Call) -> None:
         self.__top_level_attribute_stack.pop()
-        self.__in_annotation.discard(original_node)
+        self.__in_type_hint.discard(original_node)
 
     def visit_Annotation(self, node: cst.Annotation) -> Optional[bool]:
         self.__in_annotation.add(node)
@@ -726,30 +803,32 @@ class ScopeVisitor(cst.CSTVisitor):
         return False
 
     def visit_ConcatenatedString(self, node: cst.ConcatenatedString) -> Optional[bool]:
-        self._handle_string_annotation(node)
-        return False
+        return not self._handle_string_annotation(node)
 
     def _handle_string_annotation(
         self, node: Union[cst.SimpleString, cst.ConcatenatedString]
-    ) -> None:
-        if self.__in_annotation and not self.__in_ignored_subscript:
+    ) -> bool:
+        """Returns whether it successfully handled the string annotation"""
+        if (
+            self.__in_type_hint or self.__in_annotation
+        ) and not self.__in_ignored_subscript:
             value = node.evaluated_value
             if value:
                 mod = cst.parse_module(value)
                 mod.visit(self)
+                return True
+        return False
 
     def visit_Subscript(self, node: cst.Subscript) -> Optional[bool]:
-        qnames = self.scope.get_qualified_names_for(node.value)
-        if any(qn.name.startswith(("typing.", "typing_extensions.")) for qn in qnames):
-            self.__in_annotation.add(node)
-        if any(
-            qn.name in {"typing.Literal", "typing_extensions.Literal"} for qn in qnames
-        ):
+        qnames = {qn.name for qn in self.scope.get_qualified_names_for(node.value)}
+        if any(qn.startswith(("typing.", "typing_extensions.")) for qn in qnames):
+            self.__in_type_hint.add(node)
+        if "typing.Literal" in qnames or "typing_extensions.Literal" in qnames:
             self.__in_ignored_subscript.add(node)
         return True
 
     def leave_Subscript(self, original_node: cst.Subscript) -> None:
-        self.__in_annotation.discard(original_node)
+        self.__in_type_hint.discard(original_node)
         self.__in_ignored_subscript.discard(original_node)
 
     def visit_Name(self, node: cst.Name) -> Optional[bool]:
@@ -758,7 +837,14 @@ class ScopeVisitor(cst.CSTVisitor):
         if context == ExpressionContext.STORE:
             self.scope.record_assignment(node.value, node)
         elif context in (ExpressionContext.LOAD, ExpressionContext.DEL, None):
-            access = Access(node, self.scope, is_annotation=bool(self.__in_annotation))
+            access = Access(
+                node,
+                self.scope,
+                is_annotation=bool(
+                    self.__in_annotation and not self.__ignore_annotation
+                ),
+                is_type_hint=bool(self.__in_type_hint),
+            )
             self.__deferred_accesses.append(
                 (access, self.__top_level_attribute_stack[-1])
             )
@@ -817,6 +903,12 @@ class ScopeVisitor(cst.CSTVisitor):
                 statement.visit(self)
         return False
 
+    def visit_ClassDef_bases(self, node: cst.ClassDef) -> None:
+        self.__ignore_annotation += 1
+
+    def leave_ClassDef_bases(self, node: cst.ClassDef) -> None:
+        self.__ignore_annotation -= 1
+
     def visit_Global(self, node: cst.Global) -> Optional[bool]:
         for name_item in node.names:
             self.scope.record_global_overwrite(name_item.name.value)
@@ -910,7 +1002,7 @@ class ScopeVisitor(cst.CSTVisitor):
                         break
 
             scope_name_accesses[(access.scope, name)].add(access)
-            access.record_assignments(access.scope[name])
+            access.record_assignments(name)
             access.scope.record_access(name, access)
 
         for (scope, name), accesses in scope_name_accesses.items():
@@ -921,6 +1013,8 @@ class ScopeVisitor(cst.CSTVisitor):
 
     def on_leave(self, original_node: cst.CSTNode) -> None:
         self.provider.set_metadata(original_node, self.scope)
+        if isinstance(original_node, _ASSIGNMENT_LIKE_NODES):
+            self.scope._assignment_count += 1
         super().on_leave(original_node)
 
 
