@@ -111,6 +111,25 @@ class ScopeProviderTest(UnitTest):
         wrapper = MetadataWrapper(cst.parse_module("def a():\n    from b import c\n\n"))
         wrapper.visit(DependentVisitor())
 
+    def test_fstring_accesses(self) -> None:
+        m, scopes = get_scope_metadata_provider(
+            """
+            from a import b
+            f"{b}" "hello"
+            """
+        )
+        global_scope = scopes[m]
+        self.assertIsInstance(global_scope, GlobalScope)
+        global_accesses = list(global_scope.accesses)
+        self.assertEqual(len(global_accesses), 1)
+        import_node = ensure_type(
+            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
+        )
+        b_referent = list(global_accesses[0].referents)[0]
+        self.assertIsInstance(b_referent, Assignment)
+        if isinstance(b_referent, Assignment):  # for the typechecker's eyes
+            self.assertEqual(b_referent.node, import_node)
+
     @data_provider((("any",), ("True",), ("Exception",), ("__name__",)))
     def test_builtins(self, builtin: str) -> None:
         m, scopes = get_scope_metadata_provider(
@@ -1018,19 +1037,24 @@ class ScopeProviderTest(UnitTest):
     def test_annotation_access(self) -> None:
         m, scopes = get_scope_metadata_provider(
             """
-                from typing import Literal, NewType, Optional, TypeVar
-                from a import A, B, C, D, E, F, G, H
+                from typing import Literal, NewType, Optional, TypeVar, Callable, cast
+                from a import A, B, C, D, D2, E, E2, F, G, G2, H, I, J, K, K2
                 def x(a: A):
                     pass
                 def y(b: "B"):
                     pass
                 def z(c: Literal["C"]):
                     pass
-                DType = TypeVar("DType", bound=D)
-                EType = TypeVar("EType", bound="E")
+                DType = TypeVar("D2", bound=D)
+                EType = TypeVar("E2", bound="E")
                 FType = TypeVar("F")
-                GType = NewType("GType", "Optional[G]")
+                GType = NewType("G2", "Optional[G]")
                 HType = Optional["H"]
+                IType = Callable[..., I]
+
+                class Test(Generic[J]):
+                    pass
+                casted = cast("K", "K2")
             """
         )
         imp = ensure_type(
@@ -1058,13 +1082,23 @@ class ScopeProviderTest(UnitTest):
         self.assertIsInstance(assignment, Assignment)
         self.assertEqual(len(assignment.references), 1)
         references = list(assignment.references)
-        self.assertTrue(references[0].is_annotation)
+        self.assertFalse(references[0].is_annotation)
+        self.assertTrue(references[0].is_type_hint)
+
+        assignment = list(scope["D2"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 0)
 
         assignment = list(scope["E"])[0]
         self.assertIsInstance(assignment, Assignment)
         self.assertEqual(len(assignment.references), 1)
         references = list(assignment.references)
-        self.assertTrue(references[0].is_annotation)
+        self.assertFalse(references[0].is_annotation)
+        self.assertTrue(references[0].is_type_hint)
+
+        assignment = list(scope["E2"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 0)
 
         assignment = list(scope["F"])[0]
         self.assertIsInstance(assignment, Assignment)
@@ -1074,13 +1108,41 @@ class ScopeProviderTest(UnitTest):
         self.assertIsInstance(assignment, Assignment)
         self.assertEqual(len(assignment.references), 1)
         references = list(assignment.references)
-        self.assertTrue(references[0].is_annotation)
+        self.assertFalse(references[0].is_annotation)
+        self.assertTrue(references[0].is_type_hint)
+
+        assignment = list(scope["G2"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 0)
 
         assignment = list(scope["H"])[0]
         self.assertIsInstance(assignment, Assignment)
         self.assertEqual(len(assignment.references), 1)
         references = list(assignment.references)
-        self.assertTrue(references[0].is_annotation)
+        self.assertFalse(references[0].is_annotation)
+        self.assertTrue(references[0].is_type_hint)
+
+        assignment = list(scope["I"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 1)
+        references = list(assignment.references)
+        self.assertFalse(references[0].is_annotation)
+
+        assignment = list(scope["J"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 1)
+        references = list(assignment.references)
+        self.assertFalse(references[0].is_annotation)
+
+        assignment = list(scope["K"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 1)
+        references = list(assignment.references)
+        self.assertFalse(references[0].is_annotation)
+
+        assignment = list(scope["K2"])[0]
+        self.assertIsInstance(assignment, Assignment)
+        self.assertEqual(len(assignment.references), 0)
 
     def test_node_of_scopes(self) -> None:
         m, scopes = get_scope_metadata_provider(
@@ -1290,3 +1352,105 @@ class ScopeProviderTest(UnitTest):
             )
         }
         self.assertEqual(names, {"a.b.c", "a.b", "a"})
+
+    def test_ordering(self) -> None:
+        m, scopes = get_scope_metadata_provider(
+            """
+            from a import b
+            class X:
+                x = b
+                b = b
+                y = b
+            """
+        )
+        global_scope = scopes[m]
+        import_stmt = ensure_type(
+            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
+        )
+        first_assignment = list(global_scope.assignments)[0]
+        assert isinstance(first_assignment, cst.metadata.Assignment)
+        self.assertEqual(first_assignment.node, import_stmt)
+        global_refs = list(first_assignment.references)
+        self.assertEqual(len(global_refs), 2)
+        class_def = ensure_type(m.body[1], cst.ClassDef)
+        x = ensure_type(
+            ensure_type(class_def.body.body[0], cst.SimpleStatementLine).body[0],
+            cst.Assign,
+        )
+        self.assertEqual(x.value, global_refs[0].node)
+        class_b = ensure_type(
+            ensure_type(class_def.body.body[1], cst.SimpleStatementLine).body[0],
+            cst.Assign,
+        )
+        self.assertEqual(class_b.value, global_refs[1].node)
+
+        class_accesses = list(scopes[x].accesses)
+        self.assertEqual(len(class_accesses), 3)
+        self.assertIn(
+            class_b.targets[0].target,
+            [
+                ref.node
+                for acc in class_accesses
+                for ref in acc.referents
+                if isinstance(ref, Assignment)
+            ],
+        )
+        y = ensure_type(
+            ensure_type(class_def.body.body[2], cst.SimpleStatementLine).body[0],
+            cst.Assign,
+        )
+        self.assertIn(y.value, [access.node for access in class_accesses])
+
+    def test_ordering_between_scopes(self) -> None:
+        m, scopes = get_scope_metadata_provider(
+            """
+            def f(a):
+                print(a)
+                print(b)
+            a = 1
+            b = 1
+            """
+        )
+        f = cst.ensure_type(m.body[0], cst.FunctionDef)
+        a_param = f.params.params[0].name
+        a_param_assignment = list(scopes[a_param]["a"])[0]
+        a_param_refs = list(a_param_assignment.references)
+        first_print = cst.ensure_type(
+            cst.ensure_type(
+                cst.ensure_type(f.body.body[0], cst.SimpleStatementLine).body[0],
+                cst.Expr,
+            ).value,
+            cst.Call,
+        )
+        second_print = cst.ensure_type(
+            cst.ensure_type(
+                cst.ensure_type(f.body.body[1], cst.SimpleStatementLine).body[0],
+                cst.Expr,
+            ).value,
+            cst.Call,
+        )
+        self.assertEqual(
+            first_print.args[0].value,
+            a_param_refs[0].node,
+        )
+        a_global = (
+            cst.ensure_type(
+                cst.ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Assign
+            )
+            .targets[0]
+            .target
+        )
+        a_global_assignment = list(scopes[a_global]["a"])[0]
+        a_global_refs = list(a_global_assignment.references)
+        self.assertEqual(a_global_refs, [])
+        b_global = (
+            cst.ensure_type(
+                cst.ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Assign
+            )
+            .targets[0]
+            .target
+        )
+        b_global_assignment = list(scopes[b_global]["b"])[0]
+        b_global_refs = list(b_global_assignment.references)
+        self.assertEqual(len(b_global_refs), 1)
+        self.assertEqual(b_global_refs[0].node, second_print.args[0].value)
